import os
import pickle
import gzip
import pandas as pd
import joblib
from io import BytesIO
import tempfile

from azure.datalake.store import core, lib

__TEMP_STORAGE__ = os.path.join(tempfile.gettempdir(), 'ruigi')


class ADLSStorage:
    def __init__(self, parent_folder=None, **kwargs):
        os.makedirs(__TEMP_STORAGE__, exist_ok=True)
        self.parent_folder = parent_folder
        self._init(**kwargs)

    def _init(self, token=None, store_name=None, resource=None, creds=None):
        """
        Initialize Azure back-end
        """
        self.token = token or os.environ.get('ADL_TOKEN', None)
        self.store_name = store_name or os.environ.get('ADL_STORE', None)
        resource = resource or os.environ.get('ADL_RESOURCE', 'https://datalake.azure.net/')
        self.base_url = f"adl://{self.store_name}.azuredatalakestore.net"

        if self.token is None:
            if creds is None:
                creds = lib.auth(url_suffix=self.store_name, resource=resource)
            self.client = core.AzureDLFileSystem(creds, store_name=self.store_name)
        else:
            self.client = core.AzureDLFileSystem(token=token, store_name=self.store_name)

    def _upload_local(self, local_file_name, remote_file_name):
        self.client.put(local_file_name, remote_file_name)

    def _upload_buffer(self, buffer, remote_file_name):
        with self.client.open(remote_file_name, 'wb') as f:
            f.write(buffer.getvalue())

    def save(self, name, obj, format='pickle', chunk_size=None):
        """ Save file to cloud

        Args:
            name: `str`.
                Filename to be used when saving the `obj`
            obj: `obj`
                It depends on the `format` parameter.
            format: `str`
                Possible values:
                    1. `pickle`: It uses `pickle.dump` to save a gzip binary file.
                    2. `joblib`: It uses `joblib.dump` to save a BytesIO binary file.
                    3. `file`: It saves a local file sending it directly to ADLS.
                    4. `parquet`: It saves a parquet file using pandas.to_parquet.
            chunk_size: `int` default `None`
                The size of a chunk of data whenever iterating (in bytes).
                This must be a multiple of 256 KB per the API specification.
        """
        remote_file_name = '/'.join([self.parent_folder, name]) if self.parent_folder else name
        local_file_name = os.path.join(__TEMP_STORAGE__, remote_file_name.replace("/", "-"))

        if format == 'parquet':
            if not isinstance(obj, pd.DataFrame):
                # In case it is a Spark DF
                obj = obj.toPandas()
            try:
                obj.to_parquet(local_file_name)
                self._upload_local(local_file_name, remote_file_name)
            finally:
                os.remove(local_file_name)

        elif format == 'joblib':
            with BytesIO() as buffer:
                joblib.dump(obj, buffer)
                buffer.seek(0)
                self._upload_buffer(buffer, remote_file_name)
            return

        elif format == 'pickle':
            try:
                with gzip.open(local_file_name, 'wb') as f:
                    pickle.dump(obj, f, pickle.HIGHEST_PROTOCOL)
                self._upload_local(local_file_name, remote_file_name)
            finally:
                os.remove(local_file_name)

        elif format == 'file':
            self._upload_local(obj, remote_file_name)

        else:
            raise ValueError(
                "Supported formats are pickle, joblib, file or parquet")

    def load(self, name, format='pickle', columns=None, chunk_size=None):
        """
        Args:
            name: `str`.
                Filename to be load
            format: `str`
                Possible values:
                    1. `pickle`: It uses `pickle.dump` to load a gzip binary file.
                    2. `joblib`: It uses `joblib.dump` to load a BytesIO binary file.
                    3. `file`: It saves a local file sending it directly to GCS.
                    4. `parquet`: It saves a parquet file using pandas.read_parquet.
            columns: `list` default `None`
                Columns to fetch when using `parquet=True`
            chunk_size: `int` default `None`
                The size of a chunk of data whenever iterating (in bytes).
                This must be a multiple of 256 KB per the API specification.
        """
        remote_file_name = '/'.join([self.parent_folder, name]) if self.parent_folder else name
        local_file_name = os.path.join(
            __TEMP_STORAGE__, remote_file_name.replace("/", "-"))

        if format == 'file':
            with self.client.open(remote_file_name, 'rb') as fr:
                with open(local_file_name, 'wb') as fw:
                    fw.write(fr)
                return local_file_name

        elif format == 'joblib':
            with self.client.open(remote_file_name, 'rb') as fr:
                return joblib.load(fr)

        if format == 'parquet':
            with self.client.open(remote_file_name, 'rb') as fr:
                return pd.read_parquet(fr, columns=columns)

        elif format == 'pickle':
            with self.client.open(remote_file_name, 'rb') as fr:
                try:
                    # Load ZIP
                    with open(local_file_name, 'wb') as fw:
                        fw.write(fr.read())
                    # Unzip
                    with gzip.open(local_file_name, 'rb') as gr:
                        obj = pickle.load(gr)
                        return obj
                finally:
                    os.remove(local_file_name)
        else:
            raise ValueError("Supported formats are pickle, joblib, file or parquet")

    def exists(self, path):
        path = '/'.join([self.parent_folder, path]) if self.parent_folder else path
        return self.client.exists(path)

    def list_dir(self, path):
        path = '/'.join([self.parent_folder, path]) if self.parent_folder else path
        return self.client.ls(path)

    def delete(self, name):
        remote_file_name = '/'.join([self.parent_folder, name]) if self.parent_folder else name
        self.client.remove(remote_file_name)

        local_file_name = os.path.join(
            __TEMP_STORAGE__, remote_file_name.replace("/", "-"))

        if os.path.isfile(local_file_name):
            os.remove(local_file_name)



import os
import pickle
import gzip
import pandas as pd
import joblib
from io import BytesIO
import tempfile
from retry import retry
from google.resumable_media import DataCorruption
from google.api_core.exceptions import GatewayTimeout, ServiceUnavailable
from google.oauth2 import service_account
from google.cloud import storage

_RETRY_LIST = (GatewayTimeout, DataCorruption, ServiceUnavailable)
__TEMP_STORAGE__ = os.path.join(tempfile.gettempdir(), 'ruigi')

class GoogleStorage:
    def __init__(self, service_account_path, project, bucket_name, parent_folder='', ):
        os.makedirs(__TEMP_STORAGE__, exist_ok=True)

        self.project = project
        self.bucket_name = bucket_name
        self.parent_folder = parent_folder
        self._init_gcp( project, service_account_path)

    def _init_gcp(self,  project, service_account_path):
        """
        Initialize GCP back-end

        """

        #TODO: USe envs variables.
        gcp_credentials = service_account.Credentials.from_service_account_file(service_account_path)
        self.client = storage.Client(credentials=gcp_credentials, project=project)
        self.bucket = self.client.bucket(self.bucket_name)


    def save(self, name, obj, format='pickle', chunk_size=None):
        """
        Args:
            name: `str`.
                Filename to be used when saving the `obj`
            obj: `obj`
                It depends on the `format` parameter.
            format: `str`
                Possible values:
                    1. `pickle`: It uses `pickle.dump` to save a gzip binary file.
                    2. `joblib`: It uses `joblib.dump` to save a BytesIO binary file.
                    3. `file`: It saves a local file sending it directly to GCS.
                    4. `parquet`: It saves a parquet file using pandas.to_parquet.
            chunk_size: `int` default `None`
                The size of a chunk of data whenever iterating (in bytes).
                This must be a multiple of 256 KB per the API specification.
        """

        remote_file_name = os.path.join(self.parent_folder, name)
        blob = self.bucket.blob(remote_file_name, chunk_size=chunk_size)
        local_file_name = os.path.join(__TEMP_STORAGE__, remote_file_name.replace("/", "-"))

        if format=='parquet':
            if not isinstance(obj, pd.DataFrame):
                raise ValueError(f"Object to be saved as parquet must be a "
                                 f"DataFrame. Received a {type(obj)}")
            obj.to_parquet(local_file_name)
        elif format == 'joblib':
            with BytesIO() as buffer:
                joblib.dump(obj, buffer)
                buffer.seek(0)
                blob.upload_from_file(buffer)
            return
        elif format == 'pickle':
            with gzip.open(local_file_name, 'wb') as f:
                pickle.dump(obj, f, pickle.HIGHEST_PROTOCOL)
        elif format == 'file':
            local_file_name = obj
        else:
            raise ValueError("Supported formats are pickle, joblib, file or parquet")

        blob.upload_from_filename(filename=local_file_name)

    @retry(_RETRY_LIST, tries=5)
    def load(self, name, format='pickle', columns=None, chunk_size=None):
        """
        Args:
            name: `str`.
                Filename to be load
            format: `str`
                Possible values:
                    1. `pickle`: It uses `pickle.dump` to load a gzip binary file.
                    2. `joblib`: It uses `joblib.dump` to load a BytesIO binary file.
                    3. `file`: It saves a local file sending it directly to GCS.
                    4. `parquet`: It saves a parquet file using pandas.read_parquet.
            columns: `list` default `None`
                Columns to fetch when using `parquet=True`
            chunk_size: `int` default `None`
                The size of a chunk of data whenever iterating (in bytes).
                This must be a multiple of 256 KB per the API specification.
        """

        remote_file_name = os.path.join(self.parent_folder, name)
        blob = self.bucket.blob(remote_file_name, chunk_size=chunk_size)
        local_file_name = os.path.join(__TEMP_STORAGE__, remote_file_name.replace("/", "-"))

        if not blob.exists():
            raise FileNotFoundError(f'Remote file {remote_file_name} not found')

        blob.reload()
        remote_ts = blob.updated.timestamp()

        if format == 'file':
            blob.download_to_filename(local_file_name)
            return local_file_name

        else:
            buffer = BytesIO()
            blob.download_to_file(buffer)
        
        if format == 'joblib':
            return joblib.load(buffer)
        elif format == 'parquet':
            return pd.read_parquet(buffer, columns=columns)
        elif format == 'pickle':
            blob.download_to_filename(local_file_name)
            with gzip.open(local_file_name, 'rb') as f:
                return pickle.load(f)
        else:
            raise ValueError("Supported formats are pickle, joblib, file or parquet")


    def exists(self, name):

        remote_file_name = os.path.join(self.parent_folder, name)
        blob = self.bucket.blob(remote_file_name)
        return blob.exists()

    def delete(self, name):
        remote_file_name = os.path.join(self.parent_folder, name)
        blob = self.bucket.blob(remote_file_name)
        if blob.exists():
            blob.delete()

        local_file_name = os.path.join(__TEMP_STORAGE__, remote_file_name.replace("/", "-"))
        if os.path.isfile(local_file_name):
            os.remove(local_file_name)


import os
import pickle
import gzip
import pandas as pd
import joblib
from io import BytesIO
import tempfile
from retry import retry
import boto3
import botocore.exceptions

_RETRY_LIST = ()
__TEMP_STORAGE__ = os.path.join(tempfile.gettempdir(), 'ruigi')


class S3Storage:
    def __init__(self, bucket_name, aws_access_key_id=None,
                 aws_secret_access_key=None, aws_session_token=None, parent_folder='', ):
        os.makedirs(__TEMP_STORAGE__, exist_ok=True)

        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_session_token = aws_session_token
        self.bucket_name = bucket_name
        self.parent_folder = parent_folder
        self._init()

    def _init(self,):
        """
        Initialize S3 back-end

        """

        # TODO: USe envs variables.
        self.client = boto3.resource(
            's3',
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
            aws_session_token=self.aws_session_token
        )
        self.bucket = self.client.Bucket(self.bucket_name)

    def save(self, name, obj, format='pickle', chunk_size=None):
        """
        Args:
            name: `str`.
                Filename to be used when saving the `obj`
            obj: `obj`
                It depends on the `format` parameter.
            format: `str`
                Possible values:
                    1. `pickle`: It uses `pickle.dump` to save a gzip binary file.
                    2. `joblib`: It uses `joblib.dump` to save a BytesIO binary file.
                    3. `file`: It saves a local file sending it directly to GCS.
                    4. `parquet`: It saves a parquet file using pandas.to_parquet.
            chunk_size: `int` default `None`
                The size of a chunk of data whenever iterating (in bytes).
                This must be a multiple of 256 KB per the API specification.
        """

        remote_file_name = os.path.join(self.parent_folder, name)
        local_file_name = os.path.join(
            __TEMP_STORAGE__, remote_file_name.replace("/", "-"))

        if format == 'parquet':
            if not isinstance(obj, pd.DataFrame):
                raise ValueError(f"Object to be saved as parquet must be a "
                                 f"DataFrame. Received a {type(obj)}")
            obj.to_parquet(local_file_name)
        elif format == 'joblib':
            with BytesIO() as buffer:
                joblib.dump(obj, buffer)
                buffer.seek(0)
                self.bucket.upload_fileobj(buffer, remote_file_name)
            return
        elif format == 'pickle':
            with gzip.open(local_file_name, 'wb') as f:
                pickle.dump(obj, f, pickle.HIGHEST_PROTOCOL)
        elif format == 'file':
            local_file_name = obj
        else:
            raise ValueError(
                "Supported formats are pickle, joblib, file or parquet")

        self.bucket.upload_file(local_file_name, remote_file_name)

    @retry(_RETRY_LIST, tries=5)
    def load(self, name, format='pickle', columns=None, chunk_size=None):
        """
        Args:
            name: `str`.
                Filename to be load
            format: `str`
                Possible values:
                    1. `pickle`: It uses `pickle.dump` to load a gzip binary file.
                    2. `joblib`: It uses `joblib.dump` to load a BytesIO binary file.
                    3. `file`: It saves a local file sending it directly to GCS.
                    4. `parquet`: It saves a parquet file using pandas.read_parquet.
            columns: `list` default `None`
                Columns to fetch when using `parquet=True`
            chunk_size: `int` default `None`
                The size of a chunk of data whenever iterating (in bytes).
                This must be a multiple of 256 KB per the API specification.
        """

        remote_file_name = os.path.join(self.parent_folder, name)
        obj = self.bucket.Object(remote_file_name, )
        local_file_name = os.path.join(
            __TEMP_STORAGE__, remote_file_name.replace("/", "-"))

        if obj is None:
            raise FileNotFoundError(
                f'Remote file {remote_file_name} not found')

        if format == 'file':
            self.bucket.download_file(remote_file_name, local_file_name)
            return local_file_name

        else:
            buffer = BytesIO()
            self.bucket.download_fileobj(remote_file_name, buffer)

        if format == 'joblib':
            return joblib.load(buffer)
        elif format == 'parquet':
            return pd.read_parquet(buffer, columns=columns)
        elif format == 'pickle':
            self.bucket.download_file(remote_file_name, local_file_name)
            with gzip.open(local_file_name, 'rb') as f:
                return pickle.load(f)
        else:
            raise ValueError(
                "Supported formats are pickle, joblib, file or parquet")

    def exists(self, name):

        remote_file_name = os.path.join(self.parent_folder, name)
        obj = self.bucket.Object(remote_file_name)
        try:
            obj.load()
        except botocore.exceptions.ClientError:
            return False
        else:
            return True

    def delete(self, name):
        remote_file_name = os.path.join(self.parent_folder, name)
        obj = self.bucket.Object(remote_file_name)
        if obj is not None:
            obj.delete()

        local_file_name = os.path.join(
            __TEMP_STORAGE__, remote_file_name.replace("/", "-"))
        if os.path.isfile(local_file_name):
            os.remove(local_file_name)



import unittest
import os
from .. import adls_gen1
from pandas.util import testing as pd_test


class StorageTest():
    """ General Storage testing (Create, Read, Update, and Delete operations)
    """
    def get_base_test_path(self):
        return '.'

    def test_upload_and_download_file(self):
        local_file_path = 'TEST_FILE.txt'
        remote_file_path = self.get_base_test_path() + '/TEST_REMOTE_FILE.txt'

        try:
            with open(local_file_path, 'w') as fp:
                fp.write('TEST - OK')
            stg = self.get_storage()
            try:
                stg.save(remote_file_path, local_file_path, format='file')
                self.assertTrue(stg.exists(remote_file_path))
            finally:
                stg.delete(remote_file_path)
        finally:
            os.remove(local_file_path)

    def test_upload_and_download_pickle(self):
        remote_file_path = self.get_base_test_path() + '/TEST_REMOTE_FILE'

        df = pd_test.makeDataFrame()
        stg = self.get_storage()
        try:
            stg.save(remote_file_path, df, format='pickle')
            self.assertTrue(stg.exists(remote_file_path))
            df_remote = stg.load(remote_file_path, format='pickle')
            pd_test.assert_frame_equal(df, df_remote)
        finally:
            stg.delete(remote_file_path)

    def test_upload_and_download_parquet(self):
        remote_file_path = self.get_base_test_path() + '/TEST_REMOTE_FILE'

        df = pd_test.makeDataFrame()
        stg = self.get_storage()
        try:
            stg.save(remote_file_path, df, format='parquet')
            self.assertTrue(stg.exists(remote_file_path))
            df_remote = stg.load(remote_file_path, format='parquet')
            pd_test.assert_frame_equal(df, df_remote)

        finally:
            stg.delete(remote_file_path)


if __name__ == '__main__':
    unittest.main()



import luigi
import os
import pandas as pd
import joblib
import warnings
from luigi.task import flatten


class LocalTarget(luigi.LocalTarget):
    FILE_EXT = ''
    is_tmp = False

    def __init__(self, task, path=None, *args, **kwargs):
        os.makedirs(task.TARGET_DIR, exist_ok=True)
        namespace = task.get_task_namespace()
        self.task = task  # keeping task object in target object will make
        # load_inputs_params easiers
        if path is None:
            file_id = task._file_id()
            ext = '.' + self.FILE_EXT
            # this will prevent to copy all the module path to the name of the file.
            file_id = file_id.split(namespace+'.')[-1]
            path = os.path.join(task.TARGET_DIR, namespace, file_id + ext)
        super().__init__(path=path, *args, **kwargs)

    def dump_metadata(self, metadata: dict,  *args, **kwargs):
        warnings.warn("dump_metadata not implemented in LocalTarget")

    def load_metadata(self, *args, **kwargs):
        """Should return a dict."""
        warnings.warn("load_metadata not implemented in LocalTarget")
        return {}

    def remove_metadata(self, *args, **kwargs):
        warnings.warn("remove_metadata not implemented in LocalTarget")

    def get_metadata_path(self, *args, **kwargs):
        return f"{self.path}.metadata"


class CloudTarget(LocalTarget):
    """ A target that works both locally and on any storage defined in Task._storage

    """

    def __init__(self, task, *args, **kwargs):
        super().__init__(task, *args, **kwargs)

        self.has_storage = task._storage is not None

        if self.has_storage:
            self.storage = task._storage
            namespace = task.get_task_namespace()
            file_id = task._file_id()
            # this will prevent to copy all the module path to the name of the file.
            file_id = file_id.split(namespace+'.')[-1]
            # Save a copy of the local path, before modifying it: This is useful when
            self._local_path = self.path
            # the target needs to use a local path
            self.path = os.path.join(
                task.TARGET_DIR, namespace, "{}.{}".format(file_id, self.FILE_EXT))
            self.log_path = os.path.join(
                task.TARGET_DIR, namespace, "{}_log.pkl".format(file_id))

    def dump_metadata(self, metadata: dict, *args, **kwargs):
        if self.has_storage:
            assert isinstance(metadata, dict)
            self.storage.save(self.get_metadata_path(),
                              metadata, format='joblib',)
        else:
            super().dump_metadata(metadata, *args, **kwargs)

    def load_metadata(self, *args, **kwargs):
        """Should return a dict."""
        if self.has_storage:
            metadata = self.storage.load(
                self.get_metadata_path(), format='joblib',)
            assert isinstance(
                metadata, dict), f"metadata is type {type(metadata)}"
            return metadata
        else:
            return super().load_metadata(*args, **kwargs)

    def remove_metadata(self, *args, **kwargs):
        if self.has_storage:
            self.storage.delete(self.get_metadata_path())
        else:
            return super().load_metadata(*args, **kwargs)

    def load(self, *args, **kwargs):
        if self.has_storage:
            return self.load_storage(*args, **kwargs)
        return self.load_local(*args, **kwargs)

    def dump(self, *args, **kwargs):
        if self.has_storage:
            return self.dump_storage(*args, **kwargs)
        return self.dump_local(*args, **kwargs)

    def remove(self, *args, **kwargs):
        if self.has_storage:
            return self.remove_storage(*args, **kwargs)
        return self.remove_local(*args, **kwargs)

    def exists(self, *args, **kwargs):
        if self.has_storage:
            return self.exists_storage(*args, **kwargs)
        return self.exists_local(*args, **kwargs)

    def load_local(self, *args, **kwargs):
        return super().load(*args, **kwargs)

    def dump_local(self, *args, **kwargs):
        return super().dump(*args, **kwargs)

    def exists_local(self, *args, **kwargs):
        return super().exists()

    def remove_local(self, *args, **kwargs):
        return super().remove()

    def remove_storage(self, *args, **kwargs):
        self.storage.delete(self.path)

    def exists_storage(self, *args, **kwargs):
        return self.storage.exists(self.path)


class FileTarget(CloudTarget):
    """
    This target operates with filepaths.
    easy_run should return a filepath for a local temporary file. This file will be removed after been sent to Carol.
    When loading this target, the file is copied from Carol to a local file. On easy_run we receive the local filepath.
    Important note: when loading the target, its local copy will not be automatically removed.
    """

    FILE_EXT = 'file'

    # TODO NOW: Define how it should be done locally

    def load_storage(self):
        return self.storage.load(self.path, format='file')

    def dump_storage(self, tempfile_path):
        self.storage.save(self.path, tempfile_path, format='file',)
        os.remove(tempfile_path)


class PickleTarget(CloudTarget):
    FILE_EXT = 'pkl'

    def load_storage(self):
        return self.storage.load(self.path, format='joblib')

    def dump_storage(self, function_output):
        self.storage.save(self.path, function_output, format='joblib')

    def load_local(self):
        return joblib.load(self.path)

    def dump_local(self, function_output):
        os.makedirs(os.path.dirname(self.path), exist_ok=True)
        joblib.dump(function_output, self.path)


class ParquetTarget(CloudTarget):
    FILE_EXT = 'parquet'

    def load_storage(self, **kwargs):
        return self.storage.load(self.path, format='parquet', **kwargs)

    def dump_storage(self, function_output):
        self.storage.save(self.path, function_output, format='parquet',)

    def load_local(self, **kwargs):
        return pd.read_parquet(self.path, **kwargs)

    def dump_local(self, function_output):
        os.makedirs(os.path.dirname(self.path), exist_ok=True)
        function_output.to_parquet(self.path)


class KerasTarget(CloudTarget):
    FILE_EXT = 'h5'

    def load_storage(self):
        from keras.models import load_model
        local_path = self.storage.load(self.path, format='file')
        return load_model(local_path)

    def dump_storage(self, model):
        os.makedirs(os.path.dirname(self.path), exist_ok=True)
        model.save(self.path)
        self.storage.save(self.path, self.path, format='file')

    def load_local(self):
        from keras.models import load_model
        return load_model(self.path)

    def dump_local(self, model):
        os.makedirs(os.path.dirname(self.path), exist_ok=True)
        model.save(self.path)


class PytorchTarget(CloudTarget):
    FILE_EXT = 'pth'

    def load_storage(self):
        import torch
        local_path = self.storage.load(self.path, format='file')
        return torch.load(local_path)

    def dump_storage(self, model_state_dict):
        import torch
        os.makedirs(os.path.dirname(self.path), exist_ok=True)
        torch.save(model_state_dict, self.path)
        self.storage.save(self.path, self.path, format='file')

    def load_local(self):
        import torch
        return torch.load(self.path)

    def dump_local(self, model_state_dict):
        import torch
        os.makedirs(os.path.dirname(self.path), exist_ok=True)
        torch.save(model_state_dict, self.path)


class DummyTarget:

    def __init__(self, fixed_output=None, *args, **kwargs):
        self.fixed_output = fixed_output

    def exists(self):
        return True

    def complete(self):
        return all(r.complete() for r in flatten(self.requires()))

    def load(self):
        return self.fixed_output

    def dump(self, model):
        pass

    def remove(self):
        pass


class JsonTarget(CloudTarget):
    FILE_EXT = 'json'

    def load_local(self):
        return pd.read_json(self.path)

    def dump_local(self, function_output):
        # TODO: json only works for dataframe
        os.makedirs(os.path.dirname(self.path), exist_ok=True)
        function_output.to_json(self.path)



from .task import *

def f1(x):
    return x+1

def f2(x):
    return x+1

def f3(x):
    return f3(x+1)

def test_empty_task():
    class Task1(Task):
        pass
    t1 = Task1()
    assert t1

def test_if_task_generates_hash():
    class Task1(Task):
        task_function = f1
    t1 = Task1()
    h = t1.hash_version()
    assert isinstance(h,int)
    assert h !=0


def test_if_tasks_has_equal_hash():
    class Task1(Task):
        task_function = f1

    class Task2(Task):
        task_function = f2

    assert Task1().hash_version() == Task2().hash_version()

def test_if_tasks_has_different_hash_1():
    class Task1(Task):
        task_function = f1
        
    class Task3(Task):
        task_function = f3

    assert Task1().hash_version() != Task3().hash_version()

def test_if_tasks_has_different_hash_2():
    class Task1(Task):
        pass
        
    class Task2(Task):
        task_function = f1

    assert Task1().hash_version() != Task2().hash_version()

    





from setuptools import setup, find_packages

import codecs
import os
import re

here = os.path.abspath(os.path.dirname(__file__))


min_requires = [
    'luigi', "retry", 'joblib', 'pandas', 'pyarrow'
]

extras_require = {
    "dev": ['pytest', 'bumpversion', "sphinx-rtd-theme", "sphinx"],
    "azure": ['azure-datalake-store', 'adlfs']
}

extras_require["complete"] = sorted(
    {v for req in extras_require.values() for v in req}
)


def read(*parts):
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()

with open('README.md') as fobj:
    long_description = fobj.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]",
        version_file,
        re.M,
    )
    if version_match:
        return version_match.group(1)

    raise RuntimeError("Unable to find version string.")


setup(
    name='ruigi',
    setup_requires=["wheel"],
    packages=find_packages(exclude=['docs', 'doc']),
    version=find_version("ruigi/__init__.py"),
    license='Datails',
    description='Manage your pipelines easily.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Datails',
    maintainer='Datails',
    author_email='',
    url='https://github.com/datails/ruigi',
    keywords=['luigi', 'scheduling', ],
    install_requires=min_requires,
    extras_require=extras_require,
    classifiers=[
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Development Status :: 3 - Alpha',
        # Define that your audience are developers
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        "Operating System :: OS Independent",
    ],
)



from .utils import StorageTest
import unittest
from ..adls_gen1 import ADLSStorage

stg = ADLSStorage()


class TestADLGen1Storage(StorageTest, unittest.TestCase):

    def get_base_test_path(self):
        return 'test'

    def get_storage(self):
        return stg